// Background colors
var bgColor = "#363636"
var bgBorderColor = "#243300"

// Font colors
var fontColor = "#fff"

// Main colors
var mainColor1 = "#405c00"
var mainColor2 = "#97d700"
var c_section = "#565e66"

var color1 = "#23272A"
var color2 = "#2C2F33"
var color3 = "#40464e"

// Button colors
var c_button_sideBar = color1
var c_button = "#2C2F33"
var c_button_hover = "#243300"
var c_button_active = "#324700"

